import logging
from datetime import datetime
import operator

from iotile.core.hw import IOTileApp
from iotile.core.hw.update import UpdateScript
from iotile.core.dev.semver import SemanticVersion
from iotile.cloud import IOTileCloud, device_id_to_slug
from iotile_cloud.utils.basic import datetime_to_str
from typedargs.annotate import docannotate, context
from typedargs import iprint
import requests

op_map = {">=": operator.lt,
           "gteq": operator.lt,
           "<=": operator.gt,
           "lteq": operator.gt,
           ">": operator.le,
           "gt": operator.le,
           "<": operator.ge,
           "lt": operator.ge,
           "==": operator.ne,
           "eq": operator.ne
          }


def _download_ota_script(script_url, verify_server_cert):
    """Download the script from the cloud service and store to temporary file location"""

    try:
        blob = requests.get(script_url, stream=True, verify=verify_server_cert)
        return blob.content
    except Exception as e:
        iprint("Failed to download OTA script")
        iprint(e)
        return False


@context("OtaUpdater")
class OtaUpdater(IOTileApp):
    """An IOtile app that can get OTA updates from the cloud and apply them to the device.
    The primary function is a bundled "check_and_update" that

    (1) Checks the cloud for deployment requests
    (2) Grabs the most recent one for this device (if there are multiple)
    (3) Checks that the update is applicable (version)
    (4) Downloads the update script
    (5) Attempts an update
    (6) Informs the cloud of results

    Args:
        hw (HardwareManager): A HardwareManager instance connected to a
            matching device.
        app_info (tuple): The app_tag and version of the device we are
            connected to.
        os_info (tuple): The os_tag and version of the device we are
            connected to.
        device_id (int): The UUID of the device that we are connected to.
    """

    logger = logging.getLogger(__name__)

    def __init__(self, hw, app_info, os_info, device_id):
        super(OtaUpdater, self).__init__(hw, app_info, os_info, device_id)

        self.app_info = app_info
        self.os_info = os_info

        self._cloud = IOTileCloud()
        self._con = self._hw.get(8, basic=True)

        self.dev_slug = self._get_uuid()

    @classmethod
    def AppName(cls):
        return 'ota_updater'

    def _get_uuid(self):
        device_id, = self._con.rpc(0x10, 0x08, result_format="L16x")
        return device_id_to_slug(device_id)

    @docannotate
    def check_and_update(self):
        """
        Checks the cloud for an available OTA update and applies it to the device.
        Informs the cloud of the attempt afterwards
        Gives some diagnostic info to the user as process goes
        """

        script = self.check_cloud_for_script()
        if not script:
            iprint("no OTA pending")
            return

        iprint("Applying deployment " + str(script[0]))

        iprint("Downloading script")
        iprint(script[1])
        blob = _download_ota_script(script[1], self._cloud.verify_server)

        if not blob:
            iprint("Download of script failed for some reason")
            return

        iprint("Applying script")
        try:
            self._apply_ota_update(blob=blob)
            self._inform_cloud(script[0], self.dev_slug, True)
        except Exception:
            self._inform_cloud(script[0], self.dev_slug, False)
            raise

    def _check_criteria(self, criterion):

        for criteria in criterion:
            text, opr, value = criteria.split(":")
            if text == 'os_tag':
                if self.os_info[0] != int(value):
                    iprint("os_tag doesn't match: " + str(self.os_info[0]) + " != " + value)
                    return False

            elif text == 'app_tag':
                if self.app_info[0] != int(value):
                    iprint("app_tag doesn't match: " + str(self.app_info[0]) + " != " + value)
                    return False

            elif text == 'os_version':
                ver = SemanticVersion.FromString(value)
                if op_map[opr](self.os_info[1], ver):
                    iprint("os_version not compatible: " + str(self.os_info[1]) + "not" + opr + value)
                    return False

            elif text == 'app_version':
                ver = SemanticVersion.FromString(value)
                if op_map[opr](self.app_info[1], ver):
                    iprint("app_version not compatible: " + str(self.app_info[1]) + "not" + opr + value)
                    return False

            elif text == 'controller_hw_tag':
                # TODO : figure out what the check here should be
                if opr not in ('eq', '=='):
                    iprint("op needed: eq, op seen: " + opr)
                    return False
                if self._con.hardware_version() != value:
                    return False
            else:
                iprint("Unrecognized selection criteria tag : " + text)
                return None

        return True

    def check_cloud_for_script(self):
        """Checks OTA device API for latest deployment that matches the device target settings"""
        requests = self._cloud.api.ota.device(self.dev_slug).get()

        if not requests['deployments']:
            return False

        request_to_apply = None
        oldest_date = datetime.strptime('9999-12-31T00:00:00Z', "%Y-%m-%dT%H:%M:%SZ")

        for deployment in requests['deployments']:
            if (deployment['completed_on'] is None
                    and deployment['released_on'] is not None
                    and datetime.strptime(deployment['released_on'], "%Y-%m-%dT%H:%M:%SZ") < oldest_date
                    and self._check_criteria(deployment['selection_criteria'])):

                request_to_apply = deployment
                oldest_date = datetime.strptime(deployment['released_on'], "%Y-%m-%dT%H:%M:%SZ")

        if not request_to_apply:
            return False

        script = request_to_apply['script']
        deployment_id = request_to_apply['id']
        script_details = self._cloud.api.ota.script(script).file.get()
        script_url = script_details['url']
        return deployment_id, script_url

    def _apply_ota_update(self, blob):
        """"Attempt to apply script to device using the device_updater app"""

        updater = self._hw.app(name='device_updater')
        update_script = UpdateScript.FromBinary(blob)
        updater.run_script(update_script)

    def _inform_cloud(self, deployment_id, device_slug, attempt_result_bool):
        """ Inform cloud of results """

        now = datetime.now()
        attempt_str = datetime_to_str(now)

        payload = {"deployment": deployment_id,
                   "device": device_slug,
                   "attempt_successful": attempt_result_bool,
                   "last_attempt_on": attempt_str}

        self._cloud.api.ota.action.post(payload)



import pytest
from past.builtins import long
from iotile.core.exceptions import ArgumentError
from iotile.cloud.utilities import *


def test_device_slug_to_id():
    """Convert a string device slug to a number."""

    assert device_slug_to_id('d--0000-0000-0000-0010') == 0x10
    assert device_slug_to_id('d--0010') == 0x10

    with pytest.raises(ArgumentError):
        # Only 48bits allowed
        device_slug_to_id('d--1234-0000-0000-0001')

    # Null device is now acceptable
    #with pytest.raises(ArgumentError):
    #    device_slug_to_id('0000-0000-0000-0000')

    with pytest.raises(ArgumentError):
        device_slug_to_id('t--0000-0000-0000-0000')

    with pytest.raises(ArgumentError):
        device_slug_to_id('t--0000-0000-0000-0g00')

    with pytest.raises(ArgumentError):
        device_slug_to_id(0x100)


def test_device_id_to_slug():
    """Ensure we can convert device ids to slugs."""

    assert device_id_to_slug(0x10) == 'd--0000-0000-0000-0010'
    assert device_id_to_slug('d--0010') == 'd--0000-0000-0000-0010'
    assert device_id_to_slug('0010') == 'd--0000-0000-0000-0010'
    assert device_id_to_slug(0xabcd5678ef90) == 'd--0000-abcd-5678-ef90'
    assert device_id_to_slug(long(0x1234)) == 'd--0000-0000-0000-1234'

    with pytest.raises(ArgumentError):
        device_id_to_slug('string')

    with pytest.raises(ArgumentError):
        # Only 48bits allowed
        device_id_to_slug(pow(16,12))

    with pytest.raises(ArgumentError):
        device_slug_to_id(-5)


def test_fleet_id_to_slug():
    """Ensure we can convert device ids to slugs."""

    assert fleet_id_to_slug(0x10) == 'g--0000-0000-0010'
    assert fleet_id_to_slug('g--0010') == 'g--0000-0000-0010'
    assert fleet_id_to_slug('0010') == 'g--0000-0000-0010'
    assert fleet_id_to_slug(0xabcd5678ef90) == 'g--abcd-5678-ef90'
    assert fleet_id_to_slug(long(0x1234)) == 'g--0000-0000-1234'

    with pytest.raises(ArgumentError):
        fleet_id_to_slug('string')

    with pytest.raises(ArgumentError):
        # Only 48bits allowed
        fleet_id_to_slug(pow(16,12))

    with pytest.raises(ArgumentError):
        fleet_id_to_slug(-5)



"""Routines for interacting with IOTile cloud from the command line"""

from io import BytesIO
import getpass
import datetime
from collections import namedtuple
from dateutil.tz import tzutc
import dateutil.parser
import urllib3
from iotile_cloud.api.connection import Api
from iotile_cloud.api.exceptions import RestHttpBaseException, HttpNotFoundError
from iotile.core.dev.registry import ComponentRegistry
from iotile.core.dev.config import ConfigManager
from iotile.core.hw.reports import IndividualReadingReport, SignedListReport, FlexibleDictionaryReport
from iotile.core.exceptions import ArgumentError, ExternalError, DataError
from iotile.core.utilities.typedargs import context, param, return_type, annotated, type_system
from .utilities import device_id_to_slug, fleet_id_to_slug

Acknowledgement = namedtuple("Acknowledgement", ["index", "ack", "selector"])


@context("IOTileCloud")
class IOTileCloud:
    """High level routines for interacting with IOTile cloud.

    Normally, you can create one of these objects with no arguments
    and the iotile.cloud server and authentication details will
    be pulled from the ComponentRegistry.  However, you can force
    a specific domain by passing the optional domain arguments.

    If there are no stored credentials in ComponentRegistry, the
    user will be prompted for a password on the command line IF
    the session is interactive, otherwise __init__ will fail.

    Args:
        domain (str): Optional server domain.  If not specified,
            the default will be whatever is stored in the registry
        username (str): Optional username to force the user to use
            if they don't have stored credentials
    """

    DEVICE_TOKEN_TYPE = 'a-jwt'

    def __init__(self, domain=None, username=None, **kwargs):
        reg = ComponentRegistry()
        self._conf = ConfigManager()

        if domain is None:
            domain = self._conf.get('cloud:server')

        if not self.verify_server:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        _verify = self.verify_server
        self.api = Api(domain=domain, verify=_verify, **kwargs)
        self._domain = self.api.domain

        try:
            token = reg.get_config('arch:cloud_token')
            token_type = reg.get_config('arch:cloud_token_type', default='jwt')
            self.api.set_token(token, token_type=token_type)
        except ArgumentError:
            # If we are interactive, try to get the user to login for a single
            # session rather than making them call link_cloud to store a cloud token
            if type_system.interactive:
                username, password = self._prompt_user_pass(username, domain)
                ok_resp = self.api.login(email=username, password=password)

                if not ok_resp:
                    raise ExternalError("Could not login to %s as user %s" % (domain, username))
            else:
                raise ExternalError("No stored iotile cloud authentication information",
                                    suggestion='Call iotile config link_cloud with your username and password')

        self.token = self.api.token
        self.token_type = self.api.token_type

    @property
    def verify_server(self) -> bool:
        return self._conf.get('cloud:verify-server')

    def _prompt_user_pass(self, username, domain):
        if username is None:
            prompt_str = "Please enter your IOTile.cloud email: "

            username = input(prompt_str)

        prompt_str = "Please enter your IOTile.cloud password: "
        password = getpass.getpass(prompt_str)

        return username, password

    @property
    def refresh_required(self):
        return self.token_type == 'jwt'

    def _build_streamer_slug(self, device_id, streamer_id):
        idhex = "{:04x}".format(device_id)
        streamer_hex = "{:04x}".format(streamer_id)

        return "t--0000-0000-0000-{}--{}".format(idhex, streamer_hex)

    @param("device_id", "integer", desc="ID of the device that we want information about")
    @return_type("basic_dict")
    def device_info(self, device_id):
        """Query information about a device by its device id
        """

        slug = device_id_to_slug(device_id)

        try:
            dev = self.api.device(slug).get()
        except HttpNotFoundError:
            raise ArgumentError("Device does not exist in cloud database", device_id=device_id, slug=slug)

        return dev

    @param("fleet_id", "integer", desc="Id of the fleet we want to retrieve")
    @return_type("basic_dict")
    def get_fleet(self, fleet_id):
        """ Returns the devices in the given fleet."""

        api = self.api

        slug = fleet_id_to_slug(fleet_id)

        try:
            results = api.fleet(slug).devices.get()
            entries = results.get('results', [])
            return {entry.pop('device'): entry for entry in entries}
        except HttpNotFoundError:
            raise ArgumentError("Fleet does not exist in cloud database", fleet_id=fleet_id, slug=slug)

    @param("device_id", "integer", desc="Id of the device whose fleet we want to retrieve")
    @return_type("basic_dict")
    def get_whitelist(self, device_id):
        """ Returns the whitelist associated with the given device_id if any"""
        api = self.api
        slug = device_id_to_slug(device_id)
        try:
            fleets = api.fleet.get(device=slug)['results']
        except HttpNotFoundError:
            raise ExternalError("Could not find the right URL. Are fleets enabled ?")

        if not fleets:
            # This is to be expected for devices set to take data from all project, or any device.
            raise ExternalError("The device isn't in any network !")

        networks = [self.get_fleet(fleet['id']) for fleet in fleets if fleet.get('is_network', False) is True]
        networks_to_manage = [x for x in networks if x.get(slug, {}).get('is_access_point', False) is True]

        out = {}
        for network in networks_to_manage:
            out.update(network)

        # Remove ourselves from the whitelist that we are supposed to manage
        if slug in out:
            del out[slug]

        if not out:
            raise ExternalError("No device to manage in these fleets !")

        return out

    @param("max_slop", "integer", desc="Optional max time difference value")
    @return_type("bool")
    def check_time(self, max_slop=300):
        """ Check if current system time is consistent with iotile.cloud time"""

        cloud_time = self.api.session.get('{}/api/v1/server/'.format(self._domain)).json().get('now', None)
        if cloud_time is None:
            raise DataError("No date header returned from iotile cloud", domain=self._domain)

        curtime = datetime.datetime.now(tzutc())
        delta = dateutil.parser.parse(cloud_time) - curtime
        delta_secs = delta.total_seconds()

        return abs(delta_secs) < max_slop

    @param("device_id", "integer", desc="ID of the device that we want information about")
    @param("new_sg", "string", desc="The new sensor graph id that we want to load")
    @param("app_tag", "integer", desc="Optional arg to check if the device template on the cloud matches the app_tag")
    def set_sensorgraph(self, device_id, new_sg, app_tag=None):
        """The the cloud's sensor graph id that informs what kind of device this is.

        Is app_tag is passed, verify that the sensorgraph explicitly matches
        the expected app_tag by making another API call.

        Args:
            device_id (int): The id of the device that we want to change the sensorgraph for.
            new_sg (string): Name of a valid sensorgraph that you wish to set the device to
            app_tag (int): Optional. The intended app_tag of the sensorgraph will be set. If the
                app_tag passed into this function does not match the app_tag of the sensorgraph
                in iotile.cloud, raise an error.
        """
        try:
            sg = self.api.sg(new_sg).get()
        except RestHttpBaseException as exc:
            raise ExternalError("Error calling method on iotile.cloud",
                                exception=exc, response=exc.response.status_code)

        if app_tag is not None:
            if sg.get('app_tag', None) != app_tag:
                raise ArgumentError("Cloud sensorgraph record does not match app tag",
                                    value=new_sg, cloud_sg_app_tag=sg.get('app_tag', None), app_tag_set=app_tag)

        slug = device_id_to_slug(device_id)
        patch = {'sg': new_sg}

        try:
            self.api.device(slug).patch(patch)
        except RestHttpBaseException as exc:
            if exc.response.status_code == 400:
                raise ArgumentError("Error setting sensor graph, invalid value",
                                    value=new_sg, error_code=exc.response.status_code)
            else:
                raise ExternalError("Error calling method on iotile.cloud",
                                    exception=exc, response=exc.response.status_code)

    @param("device_id", "integer", desc="ID of the device that we want information about")
    @param("new_template", "string", desc="The new device template that we want to set")
    @param("os_tag", "integer", desc="Optional arg to check if the sensorgraph on the cloud matches the os_tag")
    def set_device_template(self, device_id, new_template, os_tag=None):
        """Sets the device template for the given device in iotile.cloud.

        Is os_tag is passed, verify that the device template explicitly matches
        the expected os_tag by making another API call.
        Args:
            device_id (int): The id of the device that we want to change the device template for.
            new_template (string): Name of a valid device template that you wish to set the device to
            os_tag (int): Optional. If the os_tag passed into this function does not match the
                os_tag of the device_template in iotile.cloud, raise an error.
        """
        try:
            dt = self.api.dt(new_template).get()
        except RestHttpBaseException as exc:
            raise ExternalError("Error calling method on iotile.cloud",
                                exception=exc, response=exc.response.status_code)

        if os_tag is not None:
            if dt.get('os_tag', None) != os_tag:
                raise ArgumentError("Cloud device template record does not match os tag",
                                    value=new_template, cloud_sg_os_tag=dt.get('os_tag', None), os_tag_set=os_tag)

        slug = device_id_to_slug(device_id)
        patch = {'template': new_template}

        try:
            self.api.device(slug).patch(patch, staff=1)
        except RestHttpBaseException as exc:
            if exc.response.status_code == 400:
                raise ArgumentError("Error setting device template, invalid value",
                                    value=new_template, error_code=exc.response.status_code)
            else:
                raise ExternalError("Error calling method on iotile.cloud",
                                    exception=exc, response=exc.response.status_code)

    @param("project_id", "string", desc="Optional ID of the project to download a list of devices from")
    @return_type("list(integer)")
    def device_list(self, project_id=None):
        """Download a list of all device IDs or device IDs that are members for a specific project."""

        if project_id:
            devices = self.api.device.get(project=project_id)
        else:
            devices = self.api.device.get()

        ids = [device['id'] for device in devices['results']]
        return ids

    @param("device_id", "integer", desc="ID of the device that we want to get a permanent token for")
    def impersonate_device(self, device_id):
        """Convert our token to a permanent device token.

        This function is most useful for creating virtual IOTile devices whose access to iotile.cloud
        is based on their device id, not any particular user's account.

        There are a few differences between device tokens and user tokens:
         - Device tokens never expire and don't need to be refreshed
         - Device tokens are more restricted in what they can access in IOTile.cloud than user tokens

        Args:
            device_id (int): The id of the device that we want to get a token for.
        """

        slug = device_id_to_slug(device_id)
        token_type = IOTileCloud.DEVICE_TOKEN_TYPE

        try:
            resp = self.api.device(slug).key.get(type=IOTileCloud.DEVICE_TOKEN_TYPE)
            token = resp['key']
        except RestHttpBaseException as exc:
            raise ExternalError("Error calling method on iotile.cloud",
                                exception=exc, response=exc.response.status_code)

        self.api.set_token(token, token_type=token_type)
        self.token = token
        self.token_type = token_type

        reg = ComponentRegistry()
        reg.set_config('arch:cloud_token', self.token)
        reg.set_config('arch:cloud_token_type', self.token_type)
        reg.set_config('arch:cloud_device', slug)

    @param("device_id", "integer", desc="ID of the device that we want information about")
    @param("clean", "bool", desc="Also clean old stream data for this device")
    def unclaim(self, device_id, clean=True):
        """Unclaim a device that may have previously been claimed."""

        slug = device_id_to_slug(device_id)

        payload = {'clean_streams': clean}

        try:
            self.api.device(slug).unclaim.post(payload)
        except RestHttpBaseException as exc:
            raise ExternalError("Error calling method on iotile.cloud",
                                exception=exc, response=exc.response.status_code)

    def upload_report(self, report):
        """Upload an IOTile report to the cloud.

        This function currently supports uploading the following kinds of
        reports:
            SignedListReport
            FlexibleDictionaryReport

        If you pass an instance of IndividualReadingReport, an exception will
        be thrown because IOTile.cloud does not support receiving individual
        readings.  Those are only for local use.

        The filename of the uploaded report will have an extension set based
        on the type of report that you are uploading.

        Args:
            report (IOTileReport): The report that you want to upload.  This should
                not be an IndividualReadingReport.

        Returns:
            int: The number of new readings that were accepted by the cloud as novel.
        """

        if isinstance(report, IndividualReadingReport):
            raise ArgumentError("You cannot upload IndividualReadingReport objects to iotile.cloud", report=report)

        if isinstance(report, SignedListReport):
            file_ext = ".bin"
        elif isinstance(report, FlexibleDictionaryReport):
            file_ext = ".mp"
        else:
            raise ArgumentError("Unknown report format passed to upload_report",
                                classname=report.__class__.__name__, report=report)

        timestamp = '{}'.format(report.received_time.isoformat())
        payload = {'file': ("report" + file_ext, BytesIO(report.encode()))}

        resource = self.api.streamer.report

        headers = {}
        authorization_str = '{0} {1}'.format(self.token_type, self.token)
        headers['Authorization'] = authorization_str

        resp = self.api.session.post(resource.url(), files=payload, headers=headers,
                                     params={'timestamp': timestamp})

        count = resource._process_response(resp)['count']
        return count

    def highest_acknowledged(self, device_id, streamer):
        """Get the highest acknowledged reading for a given streamer.

        Args:
            device_id (int): The device whose streamer we are querying
            streamer (int): The streamer on the device that we want info
                about.

        Returns:
            int: The highest reading id that has been acknowledged by the cloud
        """

        slug = self._build_streamer_slug(device_id, streamer)

        try:
            data = self.api.streamer(slug).get()
        except RestHttpBaseException as exc:
            raise ArgumentError("Could not get information for streamer",
                                device_id=device_id, streamer_id=streamer, slug=slug, err=str(exc))

        if 'last_id' not in data:
            raise ExternalError("Response fom the cloud did not have last_id set", response=data)

        return data['last_id']

    def device_acknowledgements(self, device_id):
        """Get all streamer acknowledgements for a device by its id.

        Args:
            device_id (int): The device we are querying

        Returns:
            list of namedtuples: A list of all acknowledgement values received from the cloud.
                The namedtuples should have index, ack and selector fields pulled from the corresponding
                record in the cloud.
        """

        slug = device_id_to_slug(device_id)

        try:
            data = self.api.streamer().get(device=slug)
        except RestHttpBaseException as exc:
            raise ArgumentError("Could not get information for streamer", device_id=device_id, slug=slug, err=str(exc))

        results = data.get('results', [])

        acknowledgements = []

        for result in results:
            acknowledgement = Acknowledgement(
                result.get("index"),
                result.get("last_id"),
                result.get("selector")
            )

            acknowledgements.append(acknowledgement)

        return acknowledgements

    @annotated
    def refresh_token(self):
        """Attempt to refresh out cloud token with iotile.cloud."""

        if self.token_type != 'jwt':
            raise DataError("Attempting to refresh a token that does not need to be refreshed",
                            token_type=self.token_type)

        conf = ConfigManager()
        domain = conf.get('cloud:server')

        url = '{}/api/v1/auth/api-jwt-refresh/'.format(domain)

        resp = self.api.session.post(url, json={'token': self.token})
        if resp.status_code != 200:
            raise ExternalError("Could not refresh token", error_code=resp.status_code)

        data = resp.json()

        # Save token that we just refreshed to the registry and update our own token
        self.token = data['token']

        reg = ComponentRegistry()
        reg.set_config('arch:cloud_token', self.token)



"""Test coverage of the cloud_uploader iotile app."""

import os
import pytest
from iotile.core.hw import HardwareManager

@pytest.fixture(scope='function')
def uploader_app():
    """A reference device with a streamer and loaded with data."""

    with HardwareManager(port="emulated:reference_1_0") as hw:
        con = hw.get(8, uuid=1, force='NRF52 ')
        sg = con.sensor_graph()

        sg.add_streamer('output 1', 'controller', False, 'hashedlist', 'telegram')
        sg.push_many('output 1', 10, 100)

        app = hw.app(name='cloud_uploader')

        yield app


def test_basic(uploader_app):
    """Make sure we can basically download reports."""

    reports = uploader_app.download(0, acknowledge=False)
    assert len(reports) == 1


def test_save(uploader_app, tmpdir):
    """Make sure we can save report files locally."""

    save_path = str(tmpdir.join('reports'))

    assert not os.path.isdir(save_path)

    uploader_app.save_locally(save_path, trigger=0)

    assert os.path.isdir(save_path)
    assert len(os.listdir(save_path)) == 1

    print(os.listdir(save_path)[0])



"""Test IOTileCloud object using a mock cloud."""

import pytest
import datetime
from dateutil.tz import tzutc

from iotile.core.exceptions import ArgumentError
from iotile.core.dev.config import ConfigManager
from iotile.cloud.config import link_cloud
from iotile.core.hw.reports import IndividualReadingReport, SignedListReport, FlexibleDictionaryReport, IOTileReading
from iotile.cloud.cloud import IOTileCloud
from iotile.cloud.cloud import Acknowledgement
from iotile.core.dev.registry import ComponentRegistry
from iotile.core.exceptions import ArgumentError, ExternalError


def test_basic_cloud(basic_cloud):
    """Make sure we can set up the mock cloud and create an IOTileCloud."""

    cloud, _proj_id, _server = basic_cloud
    assert cloud.refresh_required


def test_device_info(basic_cloud):
    """Make sure we can get device info."""

    cloud, proj_id, _server = basic_cloud

    data = cloud.device_info(1)
    assert data['id'] == 1
    assert data['project'] == proj_id
    with pytest.raises(ArgumentError):
        cloud.device_info(10)


def test_device_list(basic_cloud):
    """Make sure the device_list api works."""

    cloud, proj_id, _server = basic_cloud

    devs = cloud.device_list()
    devs_proj = cloud.device_list(project_id=proj_id)

    assert set(devs) == set([1, 2, 3, 4, 5, 6])
    assert set(devs_proj) == set([1, 2, 3, 4, 5])


def test_highest_acknowledged(basic_cloud):
    """Make sure we can get the highest_acknowledged_values."""

    cloud, proj_id, _server = basic_cloud

    assert cloud.highest_acknowledged(1, 0) == 100
    assert cloud.highest_acknowledged(1, 1) == 200

    with pytest.raises(ArgumentError):
        cloud.highest_acknowledged(6, 0)


def test_device_acknowledgements(basic_cloud):
    """Make sure we can get device acknowledgements"""

    cloud, proj_id, _server = basic_cloud

    acknowledgements = cloud.device_acknowledgements(1)

    assert len(acknowledgements) == 2

    acknowledgements.sort(key=lambda x: x.index)

    assert acknowledgements[0].ack == 100
    assert acknowledgements[1].ack == 200


def test_set_sensorgraph_basic(basic_cloud):
    """Make sure we can properly change sensorgraph"""

    cloud, proj_id, _server = basic_cloud
    data = cloud.device_info(1)
    assert data['sg'] == 'water-meter-v1-1-0'

    cloud.set_sensorgraph(1, 'water-meter-v1-1-1')
    data = cloud.device_info(1)
    assert data['sg'] == 'water-meter-v1-1-1'


def test_set_sensorgraph_check(basic_cloud):
    """Make sure we can properly change sensorgraph with app_tag checking"""
    cloud, proj_id, _server = basic_cloud
    data = cloud.device_info(1)
    assert data['sg'] == 'water-meter-v1-1-0'

    #Try changing sensorgraph while checking app_tag
    cloud.set_sensorgraph(1, 'water-meter-v1-1-1', app_tag=124)
    data = cloud.device_info(1)
    assert data['sg'] == 'water-meter-v1-1-1'

    #Try chaging sensorgraph with incorrect app_tag
    with pytest.raises(ArgumentError):
        cloud.set_sensorgraph(1, 'water-meter-v1-1-0', app_tag=0)
    data = cloud.device_info(1)
    assert data['sg'] == 'water-meter-v1-1-1'

    #Trying changing to a non-existant sg
    with pytest.raises(ExternalError):
        cloud.set_sensorgraph(1, 'water-meter-v1-1-2')
    data = cloud.device_info(1)
    assert data['sg'] == 'water-meter-v1-1-1'


def test_set_device_template_basic(basic_cloud):
    """Make sure we can properly change device template"""
    cloud, proj_id, _server = basic_cloud
    data = cloud.device_info(1)
    assert data['template'] == 'internaltestingtemplate-v0-1-0'

    cloud.set_device_template(1, 'internaltestingtemplate-v0-1-1')
    data = cloud.device_info(1)
    assert data['template'] == 'internaltestingtemplate-v0-1-1'


def test_set_device_template_check(basic_cloud):
    """Make sure we can properly change device template with os tag checking"""
    cloud, proj_id, _server = basic_cloud
    data = cloud.device_info(1)
    assert data['template'] == 'internaltestingtemplate-v0-1-0'

    #Try changing device_template while checking os_tag
    cloud.set_device_template(1, 'internaltestingtemplate-v0-1-1', os_tag=235)
    data = cloud.device_info(1)
    assert data['template'] == 'internaltestingtemplate-v0-1-1'

    #Try chaging device template with incorrect os_tag
    with pytest.raises(ArgumentError):
        cloud.set_device_template(1, 'internaltestingtemplate-v0-1-0', os_tag=0)
    data = cloud.device_info(1)
    assert data['template'] == 'internaltestingtemplate-v0-1-1'

    #Trying changing to a non-existant template
    with pytest.raises(ExternalError):
        cloud.set_device_template(1, 'internaltestingtemplate-v0-1-2')
    data = cloud.device_info(1)
    assert data['template'] == 'internaltestingtemplate-v0-1-1'


def make_sequential(iotile_id, stream, num_readings, give_ids=False, fmt="signed_list"):
    readings = []

    for i in range(0, num_readings):
        if give_ids:
            reading = IOTileReading(i, stream, i, reading_id=i+1)
        else:
            reading = IOTileReading(i, stream, i)

        readings.append(reading)

    if fmt == "signed_list":
        return SignedListReport.FromReadings(iotile_id, readings)

    return FlexibleDictionaryReport.FromReadings(iotile_id, readings, [])


def test_report_upload(basic_cloud):
    """Make sure we can properly upload reports to the cloud."""

    cloud, proj_id, server = basic_cloud

    ind_report = IndividualReadingReport.FromReadings(10, [IOTileReading(3, 1, 2)])
    signed_report = make_sequential(1, 0x5000, 10, give_ids=True)
    dict_report = make_sequential(1, 0x5000, 10, give_ids=True, fmt="flexible_dict")

    with pytest.raises(ArgumentError):
        cloud.upload_report(ind_report)

    cloud.upload_report(signed_report)
    cloud.upload_report(dict_report)



"""An IOTileApp that uploads reports from an iotile device to the cloud."""

import logging
import time
import os
import struct
from typedargs.annotate import docannotate, context
from iotile.core.exceptions import HardwareError, ArgumentError
from iotile.core.hw import IOTileApp
from iotile.core.hw.exceptions import RPCNotFoundError
from iotile.core.hw.reports import SignedListReport
from iotile.cloud import IOTileCloud, device_id_to_slug


@context("CloudUploader")
class CloudUploader(IOTileApp):
    """An IOtile app that can get reports from a device and upload them to the cloud.

    This app performs the same basic upload functionality as the IOTile Companion
    mobile application.  It has a single important method: upload.  This method
    will start a loop that:
        - acknowledges old data from the device that has safely reched iotile.cloud
        - triggers the device to send all of its data
        - waits for all data to be sent
        - uploads all reports to iotile.cloud

    Args:
        hw (HardwareManager): A HardwareManager instance connected to a
            matching device.
        app_info (tuple): The app_tag and version of the device we are
            connected to.
        os_info (tuple): The os_tag and version of the device we are
            connected to.
        device_id (int): The UUID of the device that we are connected to.
    """

    logger = logging.getLogger(__name__)

    def __init__(self, hw, app_info, os_info, device_id):
        super(CloudUploader, self).__init__(hw, app_info, os_info, device_id)

        self._cloud_obj = None
        self._con = self._hw.get(8, basic=True)

    @property
    def _cloud(self):
        if self._cloud_obj is None:
            self._cloud_obj = IOTileCloud()

        return self._cloud_obj

    @classmethod
    def AppName(cls):
        return 'cloud_uploader'

    def _get_uuid(self):
        device_id, = self._con.rpc(0x10, 0x08, result_format="L16x")
        return device_id

    def _trigger_streamer(self, index):
        error, = self._con.rpc(0x20, 0x10, index, result_format="L")

        # This error code means that the streamer did not have any new data
        if error == 0x8003801f:
            self.logger.info("We manually triggered streamer %d but it reported that there were no new readings", index)
        elif error != 0:
            raise HardwareError("Error triggering streamer", code=error, index=index)

    def _streamer_finished(self, index):
        res = self._con.rpc(0x20, 0x0a, index, result_type=(0, True))

        #Check if we got an error that means the streamer doesn't exist
        if len(res['buffer']) == 4:
            _error, = struct.unpack("<L", res['buffer'])
            return None

        comm_status, = struct.unpack("<18xBx", res['buffer'])
        return comm_status == 0

    def _wait_streamers_finished(self, timeout=60*10.0):
        start = time.time()

        while (time.time() - start) < timeout:
            for i in range(0, 16):
                self.logger.info("Waiting for streamer %d", i)
                while True:
                    status = self._streamer_finished(i)
                    if status is None:
                        self.logger.info("No streamer %d, all streamers finished", i)
                        return
                    elif status is True:
                        break

                    time.sleep(1.0)

                self.logger.info("Streamer %d finished", i)

        raise HardwareError("Device took too long to stream data", timeout_seconds=timeout)

    def _ack_streamer(self, index, value, force=False):
        if index <= 0xFF:
            error, = self._con.rpc(0x20, 0x0f, index, int(force), value, arg_format="HHL", result_format="L")

            if error == 0x8003801e:
                # This means the streamer value is older than what is already there, this is not an error
                pass
            elif error:
                raise HardwareError("Error acknowledging streamer", error_code=error, index=index, value=value)
        else:
            raise ArgumentError("Streamer index is out of bounds", index=index)

    def _wait_finished_streaming(self):
        time.sleep(1.0)
        self._wait_streamers_finished()

    def download(self, trigger=None, acknowledge=True, force=None, save=None):
        """Synchronously download all reports from the device.

        This function will:
        - acknowledge old data from the device that has safely reached iotile.cloud
          unless you pass acknowledge=False
        - trigger the device to send all of its data.  This happens automatically
          when we enable_streaming.  However, if you need to manually trigger a
          streamer, you can specify that using trigger=X where X is in the index
          of the streamer to trigger.
        - wait for all data to be received from the device.

        The only difference between this function and upload is that this function
        will return the reports as a list, rather than uploading them directly
        to iotile.cloud.

        Args:
            trigger (int): If you need to manually trigger a streamer on the device,
                you can specify its index here and it will have trigger_streamer called
                on it before we enter the upload loop.
            acknowledge (bool): If you don't want to send all cloud acknowledgements
                down to the device before enabling streaming, you can pass False.  The
                default behavior is True.
            force (dict): If you want to forcibly set the acknowledgement for a given
                streamer you can pass it in this dictionary.  The value passed will
                be used instead of whatever was received from the cloud.  It can be
                a number, which is interpreted as the value to acknowledge, or it can
                be a tuple with a number and boolean, which is interpreted as the
                force parameter.
            save (str): Optional path to save the reports that are downloaded.
                If not passed, reports will not be saved.  The path should
                point to a directory. If it does not exist, it will be created
                (as will any needed parent directories)

        Returns:
            list of IOTileReport: The list of reports received from the device.
        """

        device_id = self._get_uuid()
        slug = device_id_to_slug(device_id)

        self.logger.info("Connected to device 0x%X", device_id)

        streamer_acks = []
        if force is not None:
            for index, value in force.items():
                force_ack = False
                if isinstance(value, tuple):
                    value, force_ack = value

                streamer_acks.append((index, value, force_ack))

        if acknowledge:
            self.logger.info("Getting acknowledgements from cloud for slug %s", slug)

            resp = self._cloud.api.streamer.get(device=slug)
            acks = resp.get('results', [])
            self.logger.info("Found %d acknowledgements", len(acks))

            for ack in acks:
                index = ack['index']
                last_id = ack['last_id']

                if index <= 0xFF:
                    streamer_acks.append((index, last_id, False))

        else:
            self.logger.info("Not acknowledging readings from cloud per user request")

        for index, last_id, force in streamer_acks:
            self.logger.info("Acknowledging highest ID %d for streamer %d (force=%s)", last_id, index, force)
            self._ack_streamer(index, last_id, force=force)

        # Configure Downloader to not break up the report
        # on old pod firmware the required RPC is not implemented to allow this
        # to fail.
        self.set_report_size(allow_fail=True)  # Set to max report size
        self._hw.enable_streaming()

        if trigger is not None:
            self.logger.info("Explicitly triggering streamer %d", trigger)
            self._trigger_streamer(trigger)

        self._wait_streamers_finished()

        reports = [x for x in self._hw.iter_reports()]
        signed_reports = [x for x in reports if isinstance(x, SignedListReport)]

        self.logger.info("Received %d signed reports, ignored %d realtime reports",
                         len(signed_reports), len(reports) - len(signed_reports))

        if save is not None:
            if not os.path.exists(save):
                self.logger.debug("Creating directory to save reports: %s", save)
                os.makedirs(os.path.abspath(save), exist_ok=True)

            for report in signed_reports:
                outname = "report-{:08x}-{:04x}-{}.bin".format(report.origin, report.origin_streamer,
                                                               report.received_time.isoformat().replace(':', '_'))
                outpath = os.path.join(save, outname)

                with open(outpath, "wb") as outfile:
                    outfile.write(report.encode())

                self.logger.debug("Saved report to file %s", outname)

        return signed_reports

    @docannotate
    def upload(self, trigger=None, acknowledge=True, save=None):
        """Synchronously get all data from the device and upload it to iotile.cloud.

        This function will:
        - acknowledge old data from the device that has safely reached iotile.cloud
          unless you pass acknowledge=False
        - trigger the device to send all of its data.  This happens automatically
          when we enable_streaming.  However, if you need to manually trigger a
          streamer, you can specify that using trigger=X where X is in the index
          of the streamer to trigger.
        - wait for all data to be received from the device.
        - upload all reports to iotile.cloud securely.

        If you want to see details about what is happening, you can capture the
        logging output.

        This method will use whatever the default iotile.cloud domain and credentials
        are that are configured in your current virtualenv.

        Args:
            trigger (int): If you need to manually trigger a streamer on the device,
                you can specify its index here and it will have trigger_streamer called
                on it before we enter the upload loop.
            acknowledge (bool): If you don't want to send all cloud acknowledgements
                down to the device before enabling streaming, you can pass False.  The
                default behavior is True.
            save (str): Optional path to save the reports that are downloaded.  If not
                passed, reports will not be saved.  The path should point to a directory.
                If it does not exist, it will be created (as will any needed parent directories)
        """

        signed_reports = self.download(trigger, acknowledge, save)

        for report in signed_reports:
            self.logger.info("Uploading report with ids in (%d, %d)", report.lowest_id, report.highest_id)
            self._cloud.upload_report(report)

    @docannotate
    def save_locally(self, folder, trigger=None):
        """Synchronously get all reports from the device and save them to a local folder.

        This function **does not access iotile.cloud**, is idempotent and does
        not acknowledge any report readings either before or after downloading the reports.

        Args:
            folder (str): Optional path to save the reports that are downloaded.  If not
                passed, reports will not be saved.  The path should point to a directory.
                If it does not exist, it will be created (as will any needed parent directories)
            trigger (int): If you need to manually trigger a streamer on the device,
                you can specify its index here and it will have trigger_streamer called
                on it before we enter the upload loop.
        """

        self.download(trigger, False, save=folder)

    @docannotate
    def get_report_size(self):
        """ Gets the configured report size for a pod.

        Returns:
           int: The maximum size of a report

        """
        maxpacket, _comp1, _comp2 = self._con.rpc(0x0A, 0x06, result_format="LBB")
        return maxpacket

    @docannotate
    def set_report_size(self, size=0xFFFFFFFF, allow_fail=False):
        """ Sets and verifies the report size for a pod.

        If the POD's firmware does not support setting report size, this will
        raise an exception unless you pass ``allow_fail=True``, in which case
        it will continue with the default report size for the POD.

        Args:
            size (int): The maximum size of a report
            allow_fail (bool): Optional argument to not raise an exception if the
                POD's firmware does not support configuring report size, i.e. the
                RPC call raises RPCNotFoundError.
        """

        try:
            error, = self._con.rpc(0x0A, 0x05, size, 0, arg_format="LB", result_format="L")
        except RPCNotFoundError:
            self.logger.warning("Could not set report size because firmware was old and didn't support needed RPC 0x0a05")
            if allow_fail:
                return

            raise

        if error:
            raise HardwareError("Error setting report size.", error_code=error, size=size)

        maxpacket, _comp1, _comp2 = self._con.rpc(0x0A, 0x06, result_format="LBB")

        if maxpacket != size:
            raise HardwareError("Max Packet Size was not set as expected")
